import bpy
from bpy.types import Panel, Operator, AddonPreferences
from mathutils import Vector
from bpy.props import (
    BoolProperty,
    StringProperty,
)

bl_info = {
    "name": "keKit Modes",
    "author": "Kjell Emanuelsson",
    "category": "",
    "version": (1, 0, 0),
    "blender": (2, 80, 0),
    "location": "N-Panel/Custom",
    "warning": "",
    "description": "Mode switching module (discontinued) from keKit ",
    "doc_url": "https://kjell.gumroad.com",
}


def obj_raycast(obj, matrix, ray_origin, ray_target):
    matrix_inv = matrix.inverted()
    ray_origin_obj = matrix_inv @ ray_origin
    ray_target_obj = matrix_inv @ ray_target
    ray_direction_obj = ray_target_obj - ray_origin_obj
    success, location, normal, face_index = obj.ray_cast(ray_origin_obj, ray_direction_obj)

    if success:
        return location, normal, face_index
    else:
        return None, None, None


def mouse_raycast(context, mouse_pos, evaluated=False):
    region = context.region
    rv3d = context.region_data

    view_vector = region_2d_to_vector_3d(region, rv3d, mouse_pos)
    ray_origin = region_2d_to_origin_3d(region, rv3d, mouse_pos)
    ray_target = ray_origin + view_vector

    hit_length_squared = -1.0
    hit_obj, hit_wloc, hit_normal, hit_face = None, None, None, None
    hit, normal, face_index = None, None, None

    # cat = {'MESH', 'CURVE', 'SURFACE', 'META', 'FONT', 'HAIR', 'GPENCIL'}
    cat = ["MESH"]
    # todo: Make hit curves returnable objs w. evaluated

    if not evaluated:
        objects = [o for o in context.visible_objects]
        depsgraph = objects.copy()
        objects = [o.name for o in context.visible_objects]
    else:
        objects = [o.name for o in context.visible_objects]
        depsgraph = context.evaluated_depsgraph_get()
        depsgraph = depsgraph.object_instances

    for dup in depsgraph:
        if evaluated:
            if dup.is_instance:
                obj = dup.instance_object
                obj_mtx = dup.matrix_world.copy()
            else:
                obj = dup.object
                obj_mtx = obj.matrix_world.copy()
        else:
            obj = dup
            obj_mtx = obj.matrix_world.copy()

        if obj.type in cat and obj.name in objects:
            try:
                hit, normal, face_index = obj_raycast(obj, obj_mtx, ray_origin, ray_target)
            except RuntimeError:
                print("Raycast Failed: Unsupported object type?")
                pass

            if hit is not None:
                hit_world = obj_mtx @ hit
                length_squared = (hit_world - ray_origin).length_squared
                if hit_obj is None or length_squared < hit_length_squared:
                    hit_normal = correct_normal(obj_mtx, normal)
                    hit_wloc = hit_world
                    hit_face = face_index
                    hit_length_squared = length_squared
                    hit_obj = obj

    if hit_obj:
        return hit_obj, hit_wloc, hit_normal, hit_face
    else:
        return None, None, None, None


#
# MODULE UI
#
class UIModesModule(Panel):
    bl_idname = "UI_PT_M_MODES"
    bl_label = "Modes"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'kekit'
    # bl_options = {'HIDE_HEADER'}

    def draw(self, context):
        layout = self.layout
        row = layout.row(align=False)
        row.alignment = "CENTER"
        row.operator('view3d.ke_selmode', text="", icon="VERTEXSEL").edit_mode = "VERT"
        row.operator('view3d.ke_selmode', text="", icon="EDGESEL").edit_mode = "EDGE"
        row.operator('view3d.ke_selmode', text="", icon="FACESEL").edit_mode = "FACE"
        row.operator('view3d.ke_selmode', text="", icon="OBJECT_DATAMODE").edit_mode = "OBJECT"
        row.operator('view3d.object_mode_pie_or_toggle', text="", icon="THREE_DOTS")
        row.prop(context.preferences.addons[__name__].preferences, "selmode_mouse", text="")
        row.operator('view3d.ke_spacetoggle', text="", icon="MOUSE_MOVE")


#
# MODULE OPERATORS (MISC)
#
class KeSelectionMode(Operator):
    bl_idname = "view3d.ke_selmode"
    bl_label = "Direct Element <-> Object Mode Switch"
    bl_description = "Sets Element Mode - Direct to selection mode, also from Object Mode'\n" \
                     "Legacy Feature - Needed for 'Mouse Over Element Select Mode"

    edit_mode: bpy.props.EnumProperty(
        items=[("VERT", "Vertex Edit Mode", "", 1),
               ("EDGE", "Edge Edit Mode", "", 2),
               ("FACE", "Face Edit Mode", "", 3),
               ("OBJECT", "Object Mode", "", 4)],
        name="Edit Mode",
        default="OBJECT")

    mouse_pos = [0, 0]

    @classmethod
    def description(cls, context, properties):
        if properties.edit_mode == "VERT":
            return "Direct to Vertex Element Mode in Edit Mode'\n" \
                   "Legacy Feature - Needed for 'Mouse Over Element Select Mode"
        elif properties.edit_mode == "EDGE":
            return "Direct to Edge Element Mode in Edit Mode'\n" \
                   "Legacy Feature - Needed for 'Mouse Over Element Select Mode"
        elif properties.edit_mode == "FACE":
            return "Direct to Face Element Mode in Edit Mode'\n" \
                   "Legacy Feature - Needed for 'Mouse Over Element Select Mode"
        elif properties.edit_mode == "OBJECT":
            return "Direct to Object Mode'\n" \
                   "Legacy Feature - Needed for 'Mouse Over Element Select Mode"
        else:
            return "Set Element Mode - Direct to selection mode from Object Mode'\n" \
                   "Legacy Feature - Needed for 'Mouse Over Element Select Mode"

    @classmethod
    def poll(cls, context):
        return context.object is not None

    def invoke(self, context, event):
        self.mouse_pos[0] = event.mouse_region_x
        self.mouse_pos[1] = event.mouse_region_y
        return self.execute(context)

    def execute(self, context):
        mode = str(context.mode)
        has_componentmodes = {"MESH", "ARMATURE", "GPENCIL"}
        has_editmode_only = {"CURVE", "SURFACE", "LATTICE", "META", "HAIR", "FONT"}

        # Mouse Over select option
        if context.preferences.addons[__name__].preferences.selmode_mouse and context.space_data.type == "VIEW_3D":
            og_obj = context.object

            if mode != "OBJECT":
                bpy.ops.object.mode_set(mode="OBJECT")

            bpy.ops.view3d.select(extend=False, deselect=False, toggle=False, deselect_all=False, center=False,
                                  enumerate=False, object=False, location=self.mouse_pos)

            sel_obj = context.object

            if og_obj == sel_obj:
                if mode == "OBJECT" and sel_obj.type in (has_componentmodes | has_editmode_only):
                    bpy.ops.object.editmode_toggle()
                return {"FINISHED"}

            if mode != "OBJECT" and sel_obj.type in (has_componentmodes | has_editmode_only):
                bpy.ops.object.editmode_toggle()
            else:
                mode = "OBJECT"

        # Set selection mode
        if context.active_object is not None:
            obj = context.active_object
        else:
            obj = context.object

        if obj.type in has_componentmodes:

            if self.edit_mode != "OBJECT":

                if obj.type == "ARMATURE":
                    bpy.ops.object.posemode_toggle()

                elif obj.type == "GPENCIL":
                    if mode == "OBJECT":
                        bpy.ops.gpencil.editmode_toggle()
                    if self.edit_mode == "VERT":
                        context.scene.tool_settings.gpencil_selectmode_edit = 'POINT'
                    elif self.edit_mode == "EDGE":
                        context.scene.tool_settings.gpencil_selectmode_edit = 'STROKE'
                    elif self.edit_mode == "FACE":
                        obj.data.use_curve_edit = not obj.data.use_curve_edit
                else:
                    if mode == "OBJECT" or mode == "SCULPT":
                        bpy.ops.object.editmode_toggle()
                    bpy.ops.mesh.select_mode(type=self.edit_mode)
            else:
                if obj.type == "GPENCIL":
                    bpy.ops.gpencil.editmode_toggle()
                else:
                    bpy.ops.object.editmode_toggle()

        elif obj.type in has_editmode_only:
            bpy.ops.object.editmode_toggle()

        else:
            # print("Object does not have an Edit Mode")
            return {"CANCELLED"}

        return {"FINISHED"}


class KeSpaceToggle(Operator):
    bl_idname = "view3d.ke_spacetoggle"
    bl_label = "Space Toggle"
    bl_description = "Space Toggle: Toggle between Edit and Object modes on selected object\n" \
                     "when mouse pointer is over -nothing-"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_options = {'REGISTER'}

    mouse_pos = Vector((0, 0))

    @classmethod
    def poll(cls, context):
        return (context.object is not None and
                context.object.type in {'MESH', 'CURVE', 'SURFACE', 'META', 'FONT', 'HAIR', 'GPENCIL'})

    def invoke(self, context, event):
        self.mouse_pos[0] = event.mouse_region_x
        self.mouse_pos[1] = event.mouse_region_y
        return self.execute(context)

    def execute(self, context):
        sel_mode = str(context.mode)
        bpy.ops.object.mode_set(mode='OBJECT')
        obj, hit_wloc, hit_normal, face_index = mouse_raycast(context, self.mouse_pos)

        if not face_index:
            if sel_mode == "OBJECT":
                bpy.ops.object.mode_set(mode="EDIT")
            elif sel_mode == "EDIT_MESH":
                bpy.ops.object.mode_set(mode="OBJECT")

        return {'FINISHED'}


def update_panel(self, context):
    try:
        if "bl_rna" in UIModesModule.__dict__:
            bpy.utils.unregister_class(UIModesModule)
        UIModesModule.bl_category = context.preferences.addons[__name__].preferences.category
        bpy.utils.register_class(UIModesModule)
    except Exception as e:
        print("keModes - Panel update failed :", e)
        pass


class keModesPrefs(AddonPreferences):
    bl_idname = __name__

    category: StringProperty(
            name="Tab Category",
            description="Choose a name (category) for tab placement",
            default="kekit",
            update=update_panel
            )

    selmode_mouse: BoolProperty(name="", description="Mouse-Over Element Mode Selection for Sel.mode script",
                                default=False)

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.label(text="Tab Location (Category):")
        row.prop(self, "category", text="")
        layout.separator()


#
# MODULE REGISTRATION
#
classes = (
    keModesPrefs,
    UIModesModule,
    KeSpaceToggle,
    KeSelectionMode,
)

modules = ()


def register():
    for c in classes:
        bpy.utils.register_class(c)

    # Force Panel Udpate, for custom tab location.
    try:
        if "bl_rna" in UIModesModule.__dict__:
            bpy.utils.unregister_class(UIModesModule)
        UIModesModule.bl_category = bpy.context.preferences.addons[__name__].preferences.category
        bpy.utils.register_class(UIModesModule)
    except Exception as e:
        print("keModes panel update failed:\n", e)


def unregister():
    for c in reversed(classes):
        bpy.utils.unregister_class(c)


if __name__ == "__main__":
    register()
